////////////////////////////////////////////////////////////////////// 
// lamp/JointLikelihood.cpp 
// (c) 2000-2006 Goncalo Abecasis
// 
// This file is distributed as part of the LAMP source code package   
// and may not be redistributed in any form, without prior written    
// permission from the author. Permission is granted for you to       
// modify this file for your own personal use, but modified versions  
// must retain this copyright notice and must not be distributed.     
// 
// Permission is granted for you to use this file to compile LAMP.    
// 
// All computer programs have bugs. Use this file at your own risk.   
// 
// Wednesday March 29, 2006
// 
 
#include "JointLikelihood.h"

#include <math.h>

int JointLikelihood::CountAlleles(Mantra & m)
   {
   // To allow for disease marker linkage disequilibrium,
   // we allow two haplotypes for each marker allele,
   // one with the disease allele and one without
   return frequencies.Length();
   }

void JointLikelihood::FillPenetranceMatrix(Mantra & m)
   {
   int haplotypes     = CountAlleles(m);
   int genotypes      = haplotypes * (haplotypes + 1) / 2;

   // Allocate matrix for storing probability of observed genotypes
   // conditional on the number of matching alleles in the underlying
   // true genotype.
   penetrances.Dimension(m.n, genotypes);

   // Constant factor for likelihood
   priorLikelihood = 1.0;

   // Calculate individual specific penetrances for each individual
   for (int i = 0; i < m.n; i++)
      {
      Person & p = m.GetPerson(i);

      int observed_genotype = (marker >= 0 && p.markers[marker].isKnown()) ?
                              p.markers[marker].SequenceCoded() - 1 : -1;
      int observed_phenotype = p.affections[affection];

      if (observed_phenotype == 0 && observed_genotype == -1)
         {
         // No phenotypic data for this invidual ...
         penetrances[i].Set(1.0);
         compatibleAlleles[i].Set(1);
         continue;
         }

      compatibleAlleles[i].Zero();

      // Cycle through possible haplotype pairs j and k
      for (int geno = 0, j = 0; j < haplotypes; j++)
         for (int k = 0; k <= j; k++, geno++)
            {
            // Extract marker allele and coded genotype
            int allele_j = j / 2;
            int allele_k = k / 2;

            int genotype = (allele_j > allele_k) ?
                           allele_j * (allele_j + 1) / 2 + allele_k :
                           allele_k * (allele_k + 1) / 2 + allele_j;

            if (observed_genotype != -1 && genotype != observed_genotype ||
                frequencies[j] == 0.0 || frequencies[k] == 0.0)
               {
               penetrances[i][geno] = 0.0;
               continue;
               }

            if (observed_phenotype)
               {
               int disease_genotype = j % 2 + k % 2;

               penetrances[i][geno] = observed_phenotype == 2 ?
                                      penetrance[disease_genotype] :
                                      1.0 - penetrance[disease_genotype];
               }
            else
               penetrances[i][geno] = 1.0;

            if (penetrances[i][geno] > 0.0)
               compatibleAlleles[i][j] = compatibleAlleles[i][k] = true;
            }
      }
   }

bool JointLikelihood::isPhenotyped(Mantra & m, int i)
   {
   Person & p = m.GetPerson(i / 2);

   return marker >= 0 && p.markers[marker].isKnown() ||
          p.affections[affection] != 0;
   }

void JointLikelihood::SelectMarker(int MARKER)
   {
   int alleles = MARKER >= 0 ?
      (Pedigree::GetMarkerInfo(MARKER)->freq.Length() - 1) * 2 : 2;

   frequencies.Dimension(alleles);
   frequencies.Zero();

   marker = MARKER;
   }

void JointLikelihood::SelectAffection(int AFFECTION)
   {
   affection = AFFECTION;
   }

double JointLikelihood::GetAlleleFrequency(int founder, int allele)
   {
   return frequencies[allele];
   }

void JointLikelihood::UpdateAlleles(int founder, int carrier)
   {
   carrier = carrier / 2;

   for (int i = 0; i < frequencies.Length(); i++)
      if (compatibleAlleles[carrier][i] == false)
         HideAllele(founder, i);
   }

void JointLikelihood::UndoAlleles(int founder, int carrier)
   {
   carrier = carrier / 2;

   for (int i = 0; i < frequencies.Length(); i++)
      if (compatibleAlleles[carrier][i] == false)
         ShowAllele(founder, i);
   }

double JointLikelihood::JointProbability(Mantra & m)
   {
   // Retrieve allele frequency portion of likelihood for current state
   double lk = product.Peek();

   // Adjust likelihood according to the probability of observing each phenotype
   for (int i = 0; i < phenoList.Length(); i++)
      {
      int who = phenoList[i];

      int al1 = alleleState[m.state[who]];
      int al2 = alleleState[m.state[who + 1]];

      int genotype = al1 > al2 ? al1 * (al1 + 1) / 2 + al2:
                                 al2 * (al2 + 1) / 2 + al1;

      if (genotype >= penetrances[who / 2].Length())
         {
         lk = 0;
         genotype = 0;
         }

      lk *= penetrances[who / 2][genotype];
      }

   return lk;
   }

JointLikelihood::JointLikelihood() : penetrance(3)
   {
   allocation = 0;
   compatibleAlleles = NULL;
   }

JointLikelihood::~JointLikelihood()
   {
   if (compatibleAlleles != NULL)
      delete [] compatibleAlleles;
   }

bool JointLikelihood::ShortScoreVectors(Mantra & m)
   {
   // There is no short cut for scoring likelihoods in this case
   return false;
   }

void JointLikelihood::AllocateMemory(Mantra & m)
   {
   FuzzyInheritanceTree::AllocateMemory(m);

   if (allocation < m.n)
      {
      if (compatibleAlleles != NULL)
         delete [] compatibleAlleles;

      allocation = m.n;
      compatibleAlleles = new IntArray [m.n];
      }

   for (int i = 0; i < m.n; i++)
      compatibleAlleles[i].Dimension(frequencies.Length());
   }

void JointLikelihood::SetupAlleleList(Mantra & m)
   {
   for (int i = 0; i < frequencies.Length(); i++)
      if (frequencies[i] > 0.0)
         for (int j = 0; j < m.two_f; j++)
            ShowAllele(j, i);
   }

void JointLikelihood::CleanUpAlleleList(Mantra & m)
   {
   for (int i = 0; i < frequencies.Length(); i++)
      if (frequencies[i] > 0.0)
         for (int j = 0; j < m.two_f; j++)
            HideAllele(j, i);
   }

bool JointLikelihood::IdenticalPenetrances(int founder1, int founder2)
   {
   if (penetrances[founder1] != penetrances[founder2])
      return false;

   if (compatibleAlleles[founder1] != compatibleAlleles[founder2])
       return false;

   return true;
   }

QtlLikelihood::QtlLikelihood() : means(3)
   {
   }

QtlLikelihood::~QtlLikelihood()
   {
   }

void QtlLikelihood::SelectTrait(int TRAIT)
   {
   trait = TRAIT;
   }

#define SQUARE(x)   ((x)*(x))

void QtlLikelihood::FillPenetranceMatrix(Mantra & m)
   {
   int haplotypes     = CountAlleles(m);
   int genotypes      = haplotypes * (haplotypes + 1) / 2;

   // Allocate matrix for storing probability of observed genotypes
   // conditional on the number of matching alleles in the underlying
   // true genotype.
   penetrances.Dimension(m.n, genotypes);

   // Constant factor for likelihood
   priorLikelihood = 1.0;

   // Calculate individual specific penetrances for each individual
   for (int i = 0; i < m.n; i++)
      {
      Person & p = m.GetPerson(i);

      int observed_genotype = (marker >= 0 && p.markers[marker].isKnown()) ?
                              p.markers[marker].SequenceCoded() - 1 : -1;
      double observed_phenotype = p.traits[trait];

      if (observed_phenotype == _NAN_ && observed_genotype == -1)
         {
         // No phenotypic data for this invidual ...
         penetrances[i].Set(1.0);
         compatibleAlleles[i].Set(1);
         continue;
         }

      compatibleAlleles[i].Zero();

      // Cycle through possible haplotype pairs j and k
      for (int geno = 0, j = 0; j < haplotypes; j++)
         for (int k = 0; k <= j; k++, geno++)
            {
            // Extract marker allele and coded genotype
            int allele_j = j / 2;
            int allele_k = k / 2;

            int genotype = (allele_j > allele_k) ?
                           allele_j * (allele_j + 1) / 2 + allele_k :
                           allele_k * (allele_k + 1) / 2 + allele_j;

            if (observed_genotype != -1 && genotype != observed_genotype ||
                frequencies[j] == 0.0 || frequencies[k] == 0.0)
               {
               penetrances[i][geno] = 0.0;
               continue;
               }

            if (observed_phenotype != _NAN_)
               {
               int trait_genotype = j % 2 + k % 2;

               penetrances[i][geno] =
                  exp(-0.5 * SQUARE(observed_phenotype - means[trait_genotype]) /
                      variance) / sqrt(variance);
               }
            else
               penetrances[i][geno] = 1.0;

            if (penetrances[i][geno] > 0.0)
               compatibleAlleles[i][j] = compatibleAlleles[i][k] = true;
            }
      }
   }

bool QtlLikelihood::isPhenotyped(Mantra & m, int i)
   {
   Person & p = m.GetPerson(i / 2);

   return marker >= 0 && p.markers[marker].isKnown() ||
          p.traits[trait] != _NAN_;
   }


 
