////////////////////////////////////////////////////////////////////// 
// lamp/Main.cpp 
// (c) 2000-2006 Goncalo Abecasis
// 
// This file is distributed as part of the LAMP source code package   
// and may not be redistributed in any form, without prior written    
// permission from the author. Permission is granted for you to       
// modify this file for your own personal use, but modified versions  
// must retain this copyright notice and must not be distributed.     
// 
// Permission is granted for you to use this file to compile LAMP.    
// 
// All computer programs have bugs. Use this file at your own risk.   
// 
// Wednesday March 29, 2006
// 
 
#include "Parameters.h"
#include "Validator.h"
#include "JointModel.h"
#include "MarkerMap.h"

BEGIN_LONG_PARAMETERS(longParameters)
LONG_PARAMETER_GROUP("Disease Model")
   EXCLUSIVE_PARAMETER("additive", &JointModel::ADDITIVE)
   EXCLUSIVE_PARAMETER("dominant", &JointModel::DOMINANT)
   EXCLUSIVE_PARAMETER("recessive", &JointModel::RECESSIVE)
   EXCLUSIVE_PARAMETER("multiplicative", &JointModel::MULTIPLICATIVE)
   EXCLUSIVE_PARAMETER("free", &JointModel::FREE)
   LONG_DOUBLEPARAMETER("prevalence", &JointModel::PREVALENCE)
LONG_PARAMETER_GROUP("Constraints")
   EXCLUSIVE_PARAMETER("unrelateds-and-trios", &JointModel::UNRELATEDS)
   EXCLUSIVE_PARAMETER("sibpairs", &JointModel::SIBPAIRS)
   EXCLUSIVE_PARAMETER("none", &JointModel::NOCONSTRAINTS)
   EXCLUSIVE_PARAMETER("auto", &JointModel::AUTOMATIC)
LONG_PARAMETER_GROUP("Optimization")
   EXCLUSIVE_PARAMETER("fletcher-reeves", &JointModel::FLETCHERREEVES)
   EXCLUSIVE_PARAMETER("nelder-mead", &JointModel::NELDERMEAD)
   EXCLUSIVE_PARAMETER("stochastic", &JointModel::STOCHASTIC)
   LONG_INTPARAMETER("rounds", &JointModel::ROUNDS)
END_LONG_PARAMETERS()

#ifndef VERSION
#define VERSION "0.0.0"
#endif

int main(int argc, char ** argv)
   {
   printf("lamp -- Linkage and Association Models for Pedigree Data\n"
          "        version " VERSION " (c) 2005-2006 Goncalo Abecasis and Mingyao Li\n");

   ParameterList pl;

   String pedfile("lamp.ped"), datfile("lamp.dat");
   String frameworkMap("frame.map"), candidateMap("candidate.map");

   pl.Add(new StringParameter('p', "Pedigree File", pedfile));
   pl.Add(new StringParameter('d', "Data File", datfile));
   pl.Add(new StringParameter('f', "Framework Map", frameworkMap));
   pl.Add(new StringParameter('c', "Candidates Map", candidateMap));
   pl.Add(new LongParameters("Additional Options", longParameters));

   pl.Read(argc, argv);

   if (JointModel::NELDERMEAD == false &&
       JointModel::FLETCHERREEVES == false &&
       JointModel::STOCHASTIC == false)
       JointModel::NELDERMEAD = true;

   if (JointModel::FREE == false && JointModel::ADDITIVE == false &&
       JointModel::DOMINANT == false && JointModel::RECESSIVE == false &&
       JointModel::MULTIPLICATIVE == false)
       JointModel::FREE = true;

   if (JointModel::UNRELATEDS == false && JointModel::SIBPAIRS == false &&
       JointModel::NOCONSTRAINTS == false && JointModel::AUTOMATIC == false)
       JointModel::AUTOMATIC = true;

   pl.Status();

   Pedigree ped;

   ped.Prepare(datfile);
   ped.Load(pedfile);
   ped.LumpAlleles(0.0);
   ped.EstimateFrequencies(FREQ_ALL);

   if (ped.affectionCount == 0)
      error("No phenotype data available for analysis.\n");

   MarkerMap framework, candidates;

   framework.Load(frameworkMap, false);
   candidates.Load(candidateMap);

   Validator validator;

   validator.TrimPedigree(ped, framework, candidates);

   if (JointModel::AUTOMATIC)
      {
      printf("SELECTING MODEL CONSTRAINTS...\n"
             "==============================\n");

      switch (validator.EvaluateSample(ped, 0))
         {
         case NO_LINKAGE_INFORMATION:
            JointModel::UNRELATEDS = true;
            break;
         case SOME_LINKAGE_INFORMATION:
            JointModel::SIBPAIRS = true;
            break;
         case FULL_LINKAGE_INFORMATION:
            JointModel::NOCONSTRAINTS = true;
            break;
         }

      JointModel::AUTOMATIC = false;
      }

   JointModel joint;

   if (!joint.CheckMaps(framework, candidates))
      error("Problems with marker maps must be fixed before proceeding\n");

   joint.AllocateMemory(ped, framework);
   joint.ScoreSinglepoint(ped, framework);
   joint.ScoreConditionals(ped, framework);

   joint.Analyze(ped, framework, candidates);

   printf("\nIf you find this program useful, please cite:\n"
        "  Li, Boehnke and Abecasis (2005). Joint Modeling of Linkage\n"
        "  and Association: Identifying SNPs Responsible for a Linkage\n"
        "  Signal. Am J Hum Genet 76:934-49\n\n");
   }


 
