////////////////////////////////////////////////////////////////////// 
// lamp/ModelSummary.cpp 
// (c) 2000-2006 Goncalo Abecasis
// 
// This file is distributed as part of the LAMP source code package   
// and may not be redistributed in any form, without prior written    
// permission from the author. Permission is granted for you to       
// modify this file for your own personal use, but modified versions  
// must retain this copyright notice and must not be distributed.     
// 
// Permission is granted for you to use this file to compile LAMP.    
// 
// All computer programs have bugs. Use this file at your own risk.   
// 
// Wednesday March 29, 2006
// 
 
#include "ModelSummary.h"
#include "JointModel.h"
#include "Pedigree.h"

ModelSummary::ModelSummary()
{
    output = NULL;
}

void ModelSummary::OpenFile(const char * filename)
{
    output = fopen(filename, "wt");
    openfile = filename;
    hasOutput = false;
}

void ModelSummary::CloseFile()
{
    if (output != NULL) {
       if (hasOutput)
          printf("Closing file [%s] with estimated parameters and likelihoods\n",
                 (const char *) openfile);

       fclose(output);
       output = NULL;
    }
}

void ModelSummary::LinkageHeader()
{
    if (JointModel::UNRELATEDS)
        {
        fprintf(output, "No linkage models were fitted, because you specified\n"
                      "that your sample includes only parent-offspring trios\n"
                      "and unrelated individuals\n\n");
        fflush(output);
        return;
        }

    GenericHeader("LINKAGE", JointModel::SIBPAIRS && JointModel::FREE ?
        "To model linkage, LAMP will estimate penetrances for each of the\n"
        "three genotypes (subject to the prevalence constraint). The disease\n"
        "allele frequency is not identifiable in a sample that includes only\n"
        "affected sibpairs (or only discordant sibpairs; or only unaffected\n"
        "sibpairs) and will remain fixed. In addition to penetrance parameters,\n"
        "marker allele frequencies will be estimated when evaluating linkage\n"
        "at a marker location\n\n" :
        "To model linkage, LAMP will estimate a disease allele frequency\n"
        "and penetrances for each of the three genotypes (subject to any\n"
        "constraints you may have placed on prevalence and the genetic\n"
        "model). In addition, marker allele frequencies will be estimated\n"
        "when evaluating linkage at a marker location.\n\n"
        );
}

void ModelSummary::IndirectAssociationHeader()
{
    if (JointModel::UNRELATEDS)
        {
        fprintf(output, "No indirect association models were fitted, because\n"
                      "you specified that your sample includes only \n"
                      "parent-offspring trios and unrelated individuals\n\n"
                      "For information on allelic association, check details\n"
                      "of fitted direct association models\n\n");
        fflush(output);
        return;
        }

    GenericHeader("INDIRECT ASSOCIATION",
        "To model indirect association, LAMP will estimate frequencies\n"
        "for haplotypes including each marker allele and an (unobserved)\n"
        "disease susceptibility allele. In addition, penetrances for each\n"
        "disease locus genotype will be estimated, subject to constraints\n"
        "you may have placed on disease prevalence and genetic model.\n\n");
}

void ModelSummary::DirectAssociationHeader()
{
    GenericHeader("DIRECT ASSOCIATION",
        "Marker allele frequencies will be estimated for each allele\n"
        "using maximum likelihood. To model direct association, LAMP\n"
        "will assume that one of the  marker alleles directly influences\n"
        "disease susceptibility directly.\n\n"
        "Subject to model constraints that you specificy, three penetrances\n"
        "will be estimated: (i) one for individuals who are homozygous for\n"
        "this \"causal\" allele, (ii) another for heterozygotes, and finally\n"
        "(iii) a penetrance for individuals who carry no copies of the allele.\n\n"
        );
}

void ModelSummary::BaseModelHeader()
{
    fprintf(output, "INFORMATION ON FITTED BASE MODELS\n\n");
    fprintf(output, "This output summarizes parameter estimates and likelihoods\n"
                  "calculated by the LAMP program\n\n");
    fprintf(output, "For the base model, LAMP estimates marker allele frequencies\n"
                  "only, under the assumption of Hardy-Weinberg equilibrium\n\n");
    fprintf(output, "\n\n=============================================================\n\n");
}

void ModelSummary::GenericHeader(const char * modelType, const char * blurb)
{
    if (output == NULL) return;

    fprintf(output, "INFORMATION ON FITTED %s MODELS\n\n", modelType);
    fprintf(output, "This output summarizes parameter estimates and likelihoods\n"
                  "calculated by the LAMP program\n\n");

    fprintf(output, blurb);

    fprintf(output, "The following user specified constraints are in effect:\n"
                  "      PREVALENCE: %15g\n", JointModel::PREVALENCE);

    if (!JointModel::FREE)
       fprintf(output, "   GENETIC MODEL: %15s\n",
                JointModel::ADDITIVE ? "Additive" :
                JointModel::DOMINANT ? "Dominant" :
                JointModel::RECESSIVE ? "Recessive" :
                JointModel::MULTIPLICATIVE ? "Multiplicative" :
                "InternalError");

    fprintf(output, "\n\n=============================================================\n\n");
    fflush(output);
}

void ModelSummary::EndSection()
{
   fprintf(output, "\n=============================================================\n\n");
   fflush(output);
}

void ModelSummary::BeginSection(const char * modelType)
{
   fprintf(output, "%20s: %s\n", "TRAIT", (const char *) trait);
   fprintf(output, "%20s: %s\n", "LOCUS", (const char *) marker);
   fprintf(output, "%20s: %s\n\n", "MODEL", (const char *) modelType);

   fprintf(output, "%20s: %.4f\n", "LOG-LIKELIHOOD", loglikelihood);
   fprintf(output, "%20s: %d\n\n", "FITTED PARAMETERS", fittedParameters);

   hasOutput = true;
}

void ModelSummary::OutputBaseModel()
{
   BeginSection("BASE");

   if (markerId >= 0)
      {
      MarkerInfo * info = Pedigree::GetMarkerInfo(markerId);

      fprintf(output, "ESTIMATED ALLELE FREQUENCIES\n");
      for (int i = 0; i < frequencies.Length(); i += 2)
         fprintf(output, "  ALLELE %s : %.4f\n",
                        (const char *) info->GetAlleleLabel(i/2 + 1),
                        frequencies[i] + frequencies[i + 1]);
      fprintf(output, "\n");
      }
   else
      fprintf(output, "NO PARAMETERS ESTIMATED\n");

   EndSection();
}

void ModelSummary::OutputLinkageModel()
{
   BeginSection("LINKAGE");

   if (markerId >= 0)
      {
      MarkerInfo * info = Pedigree::GetMarkerInfo(markerId);

      fprintf(output, "ESTIMATED ALLELE FREQUENCIES\n");
      for (int i = 0; i < frequencies.Length(); i+= 2)
         fprintf(output, "  ALLELE %s : %.4f\n",
                 (const char *) info->GetAlleleLabel(i/2 + 1),
                 frequencies[i] + frequencies[i + 1]);
      fprintf(output, "\n");
      }

   OutputDiseaseModel();

   EndSection();
}

double ModelSummary::lambdaS(double daf)
{
   double prevalence = penetrances[0] * (1.0 - daf) * (1.0 - daf) +
                       penetrances[1] * (1.0 - daf) * daf * 2.0 +
                       penetrances[2] * daf * daf;

   double IBD0 = prevalence * prevalence;

   double IBD2 = Square((1.0 - daf) * penetrances[0]) +
                 2 * daf * (1.0 - daf) * penetrances[1] * penetrances[1] +
                 Square(daf * penetrances[2]);

   double IBD1 = Square((1.0 - daf) * penetrances[0]) * (1.0 - daf) +
                 Square(daf * penetrances[2]) * daf +
                 Square(1.0 - daf) * daf * penetrances[0] * penetrances[1] * 2.0 +
                 Square(daf) * (1.0 - daf) * penetrances[1] * penetrances[2] * 2.0 +
                 Square(penetrances[1]) * daf * (1.0 - daf);

   return (0.25 * IBD0 + 0.50 * IBD1 + 0.25 * IBD2) / IBD0;
}

double ModelSummary::lambdaMarker()
{
   double save0 = penetrances[0];
   double save1 = penetrances[1];
   double save2 = penetrances[2];

   double pWt1 = frequencies[1] / (frequencies[0] + frequencies[1]);
   double pWt2 = frequencies[3] / (frequencies[2] + frequencies[3]);

   penetrances[0] = pWt1 * pWt1 * penetrances[0] +
                    2 * pWt1 * (1.0 - pWt1) * penetrances[1] +
                    (1.0 - pWt1) * (1.0 - pWt1) * penetrances[2];

   penetrances[2] = pWt2 * pWt2 * penetrances[0] +
                    2 * pWt2 * (1.0 - pWt2) * penetrances[1] +
                    (1.0 - pWt2) * (1.0 - pWt2) * penetrances[2];

   penetrances[1] = pWt1 * pWt2 * penetrances[0] +
                    pWt1 * (1.0 - pWt2) * penetrances[1] +
                    (1.0 - pWt1) * pWt2 * penetrances[1] +
                    (1.0 - pWt1) * (1.0 - pWt2) * penetrances[2];

   double lambda = lambdaS(frequencies[0] + frequencies[1]);

   penetrances[0] = save0;
   penetrances[1] = save1;
   penetrances[2] = save2;

   return lambda;
}

void ModelSummary::OutputIndirectAssociationModel()
{
   BeginSection("INDIRECT ASSOCIATION");

   MarkerInfo * info = Pedigree::GetMarkerInfo(markerId);

   bool phase = penetrances[0] > penetrances[2] ? 1 : 0;

   fprintf(output, "ESTIMATED HAPLOTYPE FREQUENCIES\n\n");

   fprintf(output, "%20s     DISEASE ALLELE\n"
                   "%20s         +         -  MARGINAL\n", "", "MARKER ALLELE");

   for (int i = 0; i < frequencies.Length(); i+= 2)
      fprintf(output, "%20s %9.5f %9.5f %9.5f\n",
              (const char *) info->GetAlleleLabel(i/2 + 1),
              frequencies[i ^ phase], frequencies[i + 1 ^ phase],
              frequencies[i] + frequencies[i + 1]);
   fprintf(output, "\n");

   double daf = 0.0;
   for (int i = 0; i < frequencies.Length(); i+=2)
      daf += frequencies[i ^ phase];

   fprintf(output, "%20s %9.5f %9.5f %9.5f\n\n",
                   "MARGINAL", daf, 1.0 - daf, 1.0);

   OutputDiseaseModel();

   if (frequencies.Length() == 4)
      {
      double maf = frequencies[0] + frequencies[1];
      double d   = frequencies[0 ^ phase] - maf * daf;
      double rsq = d * d / (maf * daf * (1.0 - maf) * (1.0 - daf));

      fprintf(output, "\n");
      fprintf(output, "    RSQ(DISEASE,MARKER): %.4f\n", rsq);
      fprintf(output, "       LAMBDA_S(MARKER): %.4f\n\n", lambdaMarker());
      }

   EndSection();
}

void ModelSummary::OutputDirectAssociationModel()
{
   BeginSection("DIRECT ASSOCIATION");

   fprintf(output, "ESTIMATED ALLELE FREQUENCIES\n");
   MarkerInfo * info = Pedigree::GetMarkerInfo(markerId);

   for (int i = 0; i < frequencies.Length(); i+= 2)
      fprintf(output, "  ALLELE %s : %.4f %s\n",
              (const char *) info->GetAlleleLabel(i/2 + 1),
              frequencies[i] + frequencies[i + 1],
              frequencies[i + 1] == 0.0 ? "\"PRESUMED\" CAUSAL ALLELE" : "");
   fprintf(output, "\n");

   OutputDiseaseModel();

   EndSection();
}

void ModelSummary::OutputDiseaseModel()
{
   fprintf(output, "DISEASE LOCUS PARAMETERS\n");

   double daf = 0.0;
   for (int i = 1; i < frequencies.Length(); i+=2)
      daf += frequencies[i];

   bool phase = penetrances[0] > penetrances[2] ? 1 : 0;

   fprintf(output, "            FREQUENCY -: %.4f\n", phase ? 1.0 - daf : daf);

   fprintf(output, "         PENETRANCE +/+: %#.4g\n",   penetrances[phase ? 2 : 0]);
   fprintf(output, "         PENETRANCE +/-: %#.4g\n",   penetrances[1]);
   fprintf(output, "         PENETRANCE -/-: %#.4g\n\n", penetrances[phase ? 0 : 2]);

   fprintf(output, "              LAMDA_SIB: %.4f\n", lambdaS(daf));
   fprintf(output, "  ATTRIBUTABLE FRACTION: %.4f\n", attributableFraction(daf));
}

double ModelSummary::attributableFraction(double daf)
{
   double prevalence = penetrances[0] * (1.0 - daf) * (1.0 - daf) +
                       penetrances[1] * (1.0 - daf) * daf * 2.0 +
                       penetrances[2] * daf * daf;

   if (penetrances[0] < penetrances[2] && penetrances[0] < penetrances[1])
      return (1.0 - penetrances[0] / prevalence);

   if (penetrances[1] < penetrances[2])
      return (1.0 - penetrances[1] / prevalence);

   return (1.0 - penetrances[2] / prevalence);
}

void ModelSummary::OutputModel(int model)
{
   if (model == 0) OutputBaseModel(); else
   if (model == 1) OutputLinkageModel(); else
   if (model == 2) OutputIndirectAssociationModel(); else
   OutputDirectAssociationModel();
}


 
